package de.thdeg.greiner.superpangworld.graphics.base;

import de.thdeg.greiner.superpangworld.gameview.GameView;
import de.thdeg.greiner.superpangworld.graphics.helper.HelperValues;
import de.thdeg.greiner.superpangworld.graphics.moveable.Harpoon;
import de.thdeg.greiner.superpangworld.graphics.moveable.PlayerObject;
import de.thdeg.greiner.superpangworld.graphics.moveable.SpecialBubble;

import java.awt.geom.Point2D;
import java.util.*;

/**
 * A Bubble, which moves around on the screen and can be shot by the player.
 */
public abstract class Bubble extends CollidingGameObject implements MovingGameObject{

    /** The velocity of a normal right movement */
    public static final Point2D VELOCITY_RIGHT = new Point2D.Double(3,1);
    /** The velocity of a normal left movement */
    public static final Point2D VELOCITY_LEFT = new Point2D.Double(-3,1);
    /** The random generator */
    protected final Random random;
    /** Flag, if the bubble is in the spawning phase */
    private boolean spawning;
    /** Spawning speed */
    private int spawnSpeed;
    /** Unique id for the spawn event timer */
    protected final long spawnID;
    /** The current bubble velocity */
    protected Point2D velocity;

    /**
     * Create a bubble with default values.
     *
     * @param gameView the {@link GameView} to display the bubble
     * @param objectsToCollideWith the list of {@link CollidableGameObject} the bubble can collide with
     */
    public Bubble(GameView gameView, ArrayList<CollidableGameObject> objectsToCollideWith){
        super(gameView,objectsToCollideWith);
        rotation = 90;
        size = 10;
        width = (int) (12 * size);
        height = (int) (12 * size);
        speedInPixel = 1;

        spawnID = UUID.randomUUID().getLeastSignificantBits();
        random = new Random();
        position.setTo(random.nextInt(GameView.WIDTH-width),-width);
        spawning = true;
        spawnSpeed = 500;
        gameView.setTimer("spawn"+spawnID,"bubble"+spawnID,spawnSpeed);

        velocity = random.nextBoolean() ? VELOCITY_LEFT : VELOCITY_RIGHT;

    }

    /**
     * Create a bubble.
     *
     * @param gameView the {@link GameView} to display the bubble
     * @param objectsToCollideWith the list of {@link CollidableGameObject} the bubble can collide with
     * @param size the object size
     * @param speedInPixel the speed in pixel per tick
     * @param position the position
     * @param spawning flag, if the bubble is in the spawning phase
     * @param velocity the initial velocity of the bubble
     */
    public Bubble(GameView gameView, ArrayList<CollidableGameObject> objectsToCollideWith, double size, double speedInPixel, Position position, boolean spawning,  Point2D velocity){
        super(gameView,objectsToCollideWith);
        random = new Random();
        rotation = 90;
        this.size = size;
        this.speedInPixel = speedInPixel;
        this.position.setTo(position.x,position.y);
        this.spawning = spawning;
        this.velocity = velocity;
        spawnID = UUID.randomUUID().getLeastSignificantBits();

        width = (int) (12 * size);
        height = (int) (12 * size);

        if(spawning){
            spawnSpeed = 500;
            gameView.setTimer("spawn"+spawnID,"bubble"+spawnID,spawnSpeed);
        }


    }

    /**
     * Moves the bubble a tick further.
     */
    @Override
    public void updatePosition(){
        if(spawning){
            if(position.y >= HelperValues.FRAME_BORDER_WIDTH){
                spawning = false;
            } else if(gameView.timerExpired("spawn"+spawnID,"bubble"+spawnID)){
                position.down((width + HelperValues.FRAME_BORDER_WIDTH)/4.0);
                gameView.setTimer("spawn"+spawnID,"bubble"+spawnID,spawnSpeed);
            }
        }else{

            Point2D newLocation = new Point2D.Double(position.x,position.y);
            double speedFactor = 0.2;

            newLocation = new Point2D.Double(newLocation.getX() + (velocity.getX() * speedFactor), newLocation.getY() + (velocity.getY() * speedFactor));
            velocity = new Point2D.Double(velocity.getX() + HelperValues.BUBBLE_GRAVITY.getX(), velocity.getY() + HelperValues.BUBBLE_GRAVITY.getY());

            // Wall bounce
            if((newLocation.getX() >= GameView.WIDTH - width - HelperValues.FRAME_BORDER_WIDTH) || newLocation.getX() <= HelperValues.FRAME_BORDER_WIDTH){
                velocity = new Point2D.Double(velocity.getX() * -1.0, velocity.getY());
                if(this instanceof SpecialBubble){
                    switchSpecialBubbleState((SpecialBubble) this);
                }
            }
            // Floor bounce
            if(newLocation.getY() >= HelperValues.FRAME_WINDOW_HEIGHT + HelperValues.FRAME_BORDER_WIDTH - width){
                velocity = new Point2D.Double(velocity.getX(), velocity.getY() * -0.95);
                newLocation = new Point2D.Double(newLocation.getX(), HelperValues.FRAME_WINDOW_HEIGHT + HelperValues.FRAME_BORDER_WIDTH - width);
                if(this instanceof SpecialBubble){
                    switchSpecialBubbleState((SpecialBubble) this);
                }
            }

            position.setTo(newLocation.getX(), newLocation.getY());
        }
    }

    private void switchSpecialBubbleState(SpecialBubble specialBubble){
        specialBubble.freezeState = !specialBubble.freezeState;
        gameView.playSound("specialBounce.wav",false);
    }

    /**
     * Get the score to add when bursting the bubble.
     *
     * @return the score to add
     */
    public abstract int getBurstScore();

    /**
     * Get the size of the bubble.
     *
     * @return the size
     */
    public double getSize(){
        return this.size;
    }

    @Override
    public void reactToCollision(CollidableGameObject otherObject) {
        if (otherObject instanceof PlayerObject && !gamePlayManager.movementFroze) {
            gamePlayManager.looseLife();
        } else if (otherObject instanceof Harpoon && !spawning) {
            gamePlayManager.destroy((Harpoon) otherObject, this);
        }
    }

    /**
     * Destroys the bubble.
     */
    public abstract void destroy();
}
